/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 * 
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations
 * under the License.
 * 
 * The Original Code is Komodo code.
 * 
 * The Initial Developer of the Original Code is ActiveState Software Inc.
 * Portions created by ActiveState Software Inc are Copyright (C) 2000-2011
 * ActiveState Software Inc. All Rights Reserved.
 * 
 * Contributor(s):
 *   ActiveState Software Inc
 * 
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 * 
 * ***** END LICENSE BLOCK ***** */

/**
 * Codeintel section list control.
 */

if (typeof(ko) == 'undefined') {
    var ko = {};
}

if (typeof(ko.sectionlist)!='undefined') {
    ko.logging.getLogger('').warn("ko.sectionlist was already loaded, re-creating it.\n");
}
ko.sectionlist = {};


(function() {

    var log = ko.logging.getLogger("sectionlist");
    //log.setLevel(ko.logging.LOG_DEBUG);

    var sectionlistController;

    //---- public routines

    /**
     * Update the 'Sections' statusbar panel label.
     */
    this.update = function sectionlist_update(event, view, forceUpdate /* =false */) {
        try {
            log.debug("update()");
            if (!view) {
                if (event && "localName" in event.originalTarget && event.originalTarget.localName == "view") {
                    log.debug("update:: event.type: " + event.type);
                    view = event.originalTarget;
                } else {
                    view = ko.views.manager.currentView;
                }
            }
            var sectionlistElem = document.getElementById("statusbar-sectionlist");
            if (sectionlistElem) {
                sectionlistElem.update(view, forceUpdate);
            }
        } catch(ex) {
            log.exception(ex);
        }
    }

    this.QueryInterface = function(iid) {
        if (iid.equals(Components.interfaces.nsIObserver) ||
            iid.equals(Components.interfaces.nsISupports)) {
            return this;
        }
        throw Components.results.NS_ERROR_NO_INTERFACE;
    }
    
    this.observe = function sectionlist_observe(subject, topic, data) {
        switch (topic) {
            case 'autocomplete-did-enter-text':
                // Dev note: We can't do this with an event because this notification is sent from
                // http://mxr.mozilla.org/mozilla-central/source/toolkit/components/autocomplete/src/nsAutoCompleteController.cpp#1158
                if (subject && subject.getAttribute('anonid') == 'section-textbox') {
                    // Used by the sections list. Only observing when the menulist
                    // popup is up.
                    //
                    // This is a total HACK used to get relevant info about which
                    // section in the popup list was selected. The textbox
                    // autocomplete controller closes the popup before sending the
                    // 'ontextenter' notification that we use to handle a section
                    // selection. We need to know the popup's selected index to
                    // get the selected index (the row text isn't sufficient to
                    // uniquely identify the item).
                    var node = subject.parentNode;
                    while (node) {
                        if (node.localName == 'sectionlist') {
                            node._sectionSelectedIndex = subject.popup.selectedIndex;
                            break;
                        }
                        node = node.parentNode;
                    }
                    if (!node) {
                        log.warn("unable to locate sectionlist element for menu popup");
                    }
                }
                break;

            case 'codeintel_buffer_scanned':
                var buf = subject;
                var currentView = ko.views.manager.currentView;
                try {
                    if (currentView.koDoc.displayPath == buf.path) {
                        ko.sectionlist.update(null, currentView, true /* force update */);
                    }
                } catch(ex) {}
                break;
        }
    }

    this.initialize = function sectionlist_initialize()
    {
        log.debug("initialize()");
        try {
            ko.main.addWillCloseHandler(ko.sectionlist.finalize);
            window.addEventListener('current_view_changed', ko.sectionlist.update, false);
            window.addEventListener('current_view_language_changed', ko.sectionlist.update, false);
            window.addEventListener('current_view_linecol_changed', ko.sectionlist.update, false);
            window.addEventListener('codeintel_status_changed', ko.sectionlist.update, false);
            var obsSvc = Components.classes["@mozilla.org/observer-service;1"].
                            getService(Components.interfaces.nsIObserverService);
            obsSvc.addObserver(ko.sectionlist, 'codeintel_buffer_scanned',false);
            obsSvc.addObserver(ko.sectionlist, 'autocomplete-did-enter-text',false);
            sectionlistController = new SectionListController();
            window.controllers.appendController(sectionlistController);
        } catch(ex) {
            log.exception(ex);
        }
    }

    this.finalize = function sectionlist_finalize()
    {
        log.debug("finalize()");
        try {
            window.removeEventListener('current_view_changed', ko.sectionlist.update, false);
            window.removeEventListener('current_view_language_changed', ko.sectionlist.update, false);
            window.removeEventListener('current_view_linecol_changed', ko.sectionlist.update, false);
            window.removeEventListener('codeintel_status_changed', ko.sectionlist.update, false);
            var obsSvc = Components.classes["@mozilla.org/observer-service;1"].
                            getService(Components.interfaces.nsIObserverService);
            obsSvc.removeObserver(ko.sectionlist, 'codeintel_buffer_scanned');
            obsSvc.removeObserver(ko.sectionlist, 'autocomplete-did-enter-text');
            window.controllers.removeController(sectionlistController);
        } catch(ex) {
            log.exception(ex);
        }
    }

    //---- The controller for the sectionlist commands.

    function SectionListController() {
    }
    // The following two lines ensure proper inheritance (see Flanagan, p. 144).
    SectionListController.prototype = new xtk.Controller();
    SectionListController.prototype.constructor = SectionListController;

    SectionListController.prototype.is_cmd_showSectionList_supported = function() {
        return ko.prefs.getBooleanPref("codeintel_enabled");
    }

    SectionListController.prototype.is_cmd_showSectionList_enabled = function() {
        var view = ko.views.manager.currentView;
        return view && view.getAttribute("type") == "editor";
    }

    SectionListController.prototype.do_cmd_showSectionList = function() {
        var view = ko.views.manager.currentView;
        if (view) {
            var sectionlistElem = document.getElementById("statusbar-sectionlist");
            if (sectionlistElem) {
                sectionlistElem.open();
            }
        }
    }

}).apply(ko.sectionlist);

window.addEventListener("komodo-ui-started", ko.sectionlist.initialize, false);
